/*******************************************************************************
*
*   testing GLX_EXT_texture_from_pixmap with pure X11/GLX program
*
*   This program is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License version 3,
*   as published by the Free Software Foundation.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*   author: Mirco Müller <macslow@bangang.de>
*
*******************************************************************************/

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <X11/keysymdef.h>
#include <X11/extensions/Xcomposite.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#define Z_NEAR  1.0f
#define Z_FAR  50.0f
#define FOVY   20.0f

typedef struct _PixmapTexture
{
	Display*  pDisplay;
	Window    window;
	Pixmap    pixmap;
	GLXPixmap glxpixmap;
	GLenum    iTarget;
	GLuint    uiTextureId;
	GLint     iWidth;
	GLint     iHeight;
	GLfloat   fTop;
	GLfloat   fBottom;
} PixmapTexture;

typedef GLvoid
(*glXBindTexImageEXTFunc) (Display*,
			   GLXDrawable,
			   int,
			   const int*);

typedef GLvoid
(*glXReleaseTexImageEXTFunc) (Display*,
			      GLXDrawable,
			      int);

glXBindTexImageEXTFunc    glXBindTexImageEXT    = 0;
glXReleaseTexImageEXTFunc glXReleaseTexImageEXT = 0;

static PixmapTexture*
pixmap_texture_new (Display* pDisplay,
		    Window   window)
{
	PixmapTexture*    pPixmapTexture = NULL;
	int               aiGLXPixmapAttribs[] = {
				GLX_TEXTURE_TARGET_EXT,
				GLX_TEXTURE_RECTANGLE_EXT,
				GLX_TEXTURE_FORMAT_EXT,
				GLX_TEXTURE_FORMAT_RGB_EXT,
				GLX_MIPMAP_TEXTURE_EXT,
				GL_TRUE,
				None };
	Bool              bInvertedY;
	XWindowAttributes attribs;
	unsigned int      uiValue;
	GLXFBConfig*      pConfig = NULL;
	int               aiAttribs[3];
	int               iNumOfElements;

	pPixmapTexture = (PixmapTexture*) calloc (1, sizeof (PixmapTexture));
	if (!pPixmapTexture)
		return NULL;

	glXQueryDrawable (pDisplay, window, GLX_FBCONFIG_ID, &uiValue);
	aiAttribs[0] = GLX_FBCONFIG_ID;
	aiAttribs[1] = (int) uiValue;
	aiAttribs[2] = None;
	pConfig = glXChooseFBConfig (pDisplay, 0, aiAttribs, &iNumOfElements);

	pPixmapTexture->pDisplay  = pDisplay;
	pPixmapTexture->window    = window;
	pPixmapTexture->pixmap    = XCompositeNameWindowPixmap (pDisplay,
								window);
	pPixmapTexture->glxpixmap = glXCreatePixmap (pDisplay,
						     *pConfig,
						     pPixmapTexture->pixmap,
						     aiGLXPixmapAttribs);

	pPixmapTexture->iTarget   = GL_TEXTURE_RECTANGLE_ARB;
	glGenTextures (1, &pPixmapTexture->uiTextureId);

	XGetWindowAttributes (pDisplay,
			      window,
			      &attribs);
	pPixmapTexture->iWidth  = attribs.width;
	pPixmapTexture->iHeight = attribs.height;

	glXGetFBConfigAttrib (pDisplay,
			      *pConfig,
                              GLX_Y_INVERTED_EXT,
                              &bInvertedY);
	if (bInvertedY)
	{
		pPixmapTexture->fTop = 0.0f;
		pPixmapTexture->fBottom = 1.0f * (float)pPixmapTexture->iHeight;
	}
	else
	{
		pPixmapTexture->fTop = 1.0f * (float) pPixmapTexture->iHeight;
		pPixmapTexture->fBottom = 0.0f;
	}

	return pPixmapTexture;
}

static void
pixmap_texture_delete (PixmapTexture* pPixmapTexture)
{
	if (!pPixmapTexture)
		return;

	glXDestroyPixmap (pPixmapTexture->pDisplay, pPixmapTexture->glxpixmap);
	XFreePixmap (pPixmapTexture->pDisplay, pPixmapTexture->pixmap);
	glDeleteTextures (1, &pPixmapTexture->uiTextureId);
	free ((void*) pPixmapTexture);
}

static void
init_tfp (Display* pDisplay,
	  int      iScreen)
{
	const char* pcDummy = NULL;

	pcDummy = glXGetClientString (pDisplay, GLX_EXTENSIONS);

	if (strstr (pcDummy, "GLX_EXT_texture_from_pixmap") != NULL)
	{
		glXBindTexImageEXT = (glXBindTexImageEXTFunc)
			glXGetProcAddress ((GLubyte*) "glXBindTexImageEXT");

		glXReleaseTexImageEXT = (glXReleaseTexImageEXTFunc)
			glXGetProcAddress ((GLubyte*) "glXReleaseTexImageEXT");
	}
	else
		printf ("Unable to find texture-from-pixmap support!\n");
}

static void
change_projection (Bool    bParallel,
		   GLsizei iWidth,
		   GLsizei iHeight)
{
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	if (bParallel)
		glOrtho ((GLdouble) -iWidth / 2, (GLdouble) iWidth / 2,
			 (GLdouble) -iHeight / 2, (GLdouble) iHeight / 2,
			 Z_NEAR, Z_FAR);
	else
		gluPerspective (2.0f * FOVY,
				(GLdouble) iWidth / (GLdouble) iHeight,
				Z_NEAR,
				Z_FAR);

	glMatrixMode (GL_MODELVIEW);
}

static void
draw (Display*       pDisplay,
      Window         window,
      int            iWidth,
      int            iHeight,
      PixmapTexture* pPixmapTexture)
{
	static GLfloat fAngle = 0.0f;

	glClear (GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	change_projection (GL_TRUE, iWidth, iHeight);

	glPushMatrix ();
	glRotatef (fAngle, 0.0f, 0.0f, 1.0f);
	glColor3f (0.0f, 1.0f, 0.0f);
	glRectf (-250.0f, -30.0f, 250.0f, 30.0f);
	glPopMatrix ();
	fAngle += 1.0f;

	change_projection (False, iWidth, iHeight);

	glBindTexture (pPixmapTexture->iTarget,
		       pPixmapTexture->uiTextureId);
	glXBindTexImageEXT (pPixmapTexture->pDisplay,
			    pPixmapTexture->glxpixmap,
			    GLX_FRONT_LEFT_EXT,
			    NULL);
	glTexParameteri (pPixmapTexture->iTarget,
			 GL_TEXTURE_MIN_FILTER,
			 GL_LINEAR);
	glTexParameteri (pPixmapTexture->iTarget,
			 GL_TEXTURE_MAG_FILTER,
			 GL_LINEAR);

	glEnable (GL_DEPTH_TEST);
	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);

	glPushMatrix ();
	glTranslatef ( 1.2f * sin (0.1f * fAngle),
		      -1.5f * cos (0.05f * fAngle),
		      -6.0f + cos (0.15f * fAngle));
	glRotatef (-1.0f * fAngle, 1.0f, 0.0f, 0.0f);
	glRotatef (-0.5f * fAngle, 0.0f, 1.0f, 0.0f);
	glRotatef ( 2.0f * fAngle, 0.0f, 0.0f, 1.0f);
	glBegin (GL_QUADS);
	glNormal3f ( 0.0f, 0.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f, 1.0f, 1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f, 1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f,-1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f,-1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f, 1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f,-1.0f,-1.0f);
	glNormal3f ( 1.0f, 0.0f, 0.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f ( 1.0f,-1.0f, 1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f ( 1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f,-1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f (-1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f, 1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f (-1.0f, 1.0f,-1.0f);
	glNormal3f ( 0.0f, 1.0f, 0.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f ( 1.0f, 1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f, 1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f (-1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f,-1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f ( 1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f (-1.0f,-1.0f, 1.0f);
	glEnd ();
	glPopMatrix ();

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_LIGHTING);
	glDisable (GL_LIGHT0);

	glXReleaseTexImageEXT (pPixmapTexture->pDisplay,
			       pPixmapTexture->glxpixmap,
			       GLX_FRONT_LEFT_EXT);

	glXSwapBuffers (pDisplay, window);
}

static void
resize (int iWidth,
	int iHeight)
{
	glViewport (0, 0, iWidth, iHeight);
	change_projection (GL_TRUE, iWidth, iHeight);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

static Window
create_window (Display* pDisplay,
	       int      iWidth,
	       int      iHeight)
{
	int aiAttribs[] = {GLX_RGBA,
			   GLX_RED_SIZE, 8,
			   GLX_GREEN_SIZE, 8,
			   GLX_BLUE_SIZE, 8,
			   GLX_DEPTH_SIZE, 24,
			   GLX_DOUBLEBUFFER,
			   None};
	int                  iScreen;
	XSetWindowAttributes winAttribs;
	unsigned long        ulMask;
	Window               rootWindow;
	Window               window;
	XVisualInfo*         pVisInfo;
	GLXContext           context;

	iScreen = DefaultScreen (pDisplay);
	rootWindow = RootWindow (pDisplay, iScreen);

	pVisInfo = glXChooseVisual (pDisplay, iScreen, aiAttribs);
	if (!pVisInfo)
	{
		printf ("Error: couldn't get an RGB, Double-buffered visual\n");
		exit (1);
	}

	/* window attributes */
	winAttribs.background_pixel = 0;
	winAttribs.border_pixel     = 0;
	winAttribs.colormap         = XCreateColormap (pDisplay,
						       rootWindow,
						       pVisInfo->visual,
						       AllocNone);
	winAttribs.event_mask       = StructureNotifyMask |
				      KeyPressMask;
	ulMask                      = CWBackPixel |
				      CWBorderPixel |
				      CWColormap |
				      CWEventMask;

	window = XCreateWindow (pDisplay,
				rootWindow,
				0,
				0,
				iWidth,
				iHeight,
				0,
				pVisInfo->depth,
				InputOutput,
				pVisInfo->visual,
				ulMask,
				&winAttribs);

	context = glXCreateContext (pDisplay, pVisInfo, NULL, False);
	if (!context)
	{
		printf ("Error: glXCreateContext failed\n");
		exit (1);
	}

	glXMakeCurrent (pDisplay, window, context);

	return window;
}


static void
event_loop (Display* pDisplay,
	    Window   window)
{
	XEvent         event;
	Bool           bKeepGoing     = True;
	Bool           bGotEvent      = False;
	int            iOldWidth      = 0;
	int            iOldHeight     = 0;
	int            iNewWidth      = 0;
	int            iNewHeight     = 0;
	PixmapTexture* pPixmapTexture = NULL;

	while (bKeepGoing)
	{
		bGotEvent = XCheckWindowEvent (pDisplay,
					       window,
					       KeyPressMask |
					       StructureNotifyMask,
					       &event);

		if (bGotEvent)
		{
			switch (event.type)
			{
				case KeyPress :
					if (XLookupKeysym (&event.xkey,
							   0) == XK_Escape)
						bKeepGoing = False;
					if (XLookupKeysym (&event.xkey,
							   0) == XK_q)
						bKeepGoing = False;
				break;

				case ConfigureNotify :
					iNewWidth  = event.xconfigure.width;
					iNewHeight = event.xconfigure.height;
					if (iNewWidth != iOldWidth ||
					    iNewHeight != iNewHeight)
					{
						resize (iNewWidth, iNewHeight);

						pixmap_texture_delete (
							pPixmapTexture);
						pPixmapTexture =
						   pixmap_texture_new (pDisplay,
								       window);

						iOldWidth  = iNewWidth;
						iOldHeight = iNewHeight;
					}
				break;
			}
		}
		else
		{
			draw (pDisplay,
			      window,
			      iOldWidth,
			      iOldHeight,
			      pPixmapTexture);
			usleep (20000);
		}
	}

	pixmap_texture_delete (pPixmapTexture);
}

int
main (int    argc,
      char** argv)
{
	Display *pDisplay;
	Window window;
	int iEventBase = 0;
	int iErrorBase = 0;
	int iMajor = 0;
	int iMinor = 0;

	pDisplay = XOpenDisplay (NULL);

	/* make operations synchronous */
	XSynchronize (pDisplay, True);

	/* check for Xcomposite-extension */
	if (!XCompositeQueryExtension (pDisplay, &iEventBase, &iErrorBase))
	{
		XCloseDisplay (pDisplay);
		printf ("Xcomposite-extension not found");
		return 1;
	}

	/* see which version of Xcomposite is available */
	XCompositeQueryVersion (pDisplay, &iMajor, &iMinor);
	if (iMajor == 0 && iMinor < 4)
	{
		XCloseDisplay (pDisplay);
		printf ("At least Xcomposite 0.4 needed\n");
		return 2;
	}
	printf ("Found Xcomposite %d.%d\n", iMajor, iMinor);

	window = create_window (pDisplay, 300, 300);

	glClearColor (1.0f, 0.5f, 0.25f, 1.0f);
	glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, 1);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_RECTANGLE_ARB);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	gluLookAt (0.0f, 0.0f, Z_NEAR,
		   0.0f, 0.0f, 0.0f,
		   0.0f, 1.0f, 0.0f);
	glTranslatef (0.0f, 0.0f, -Z_NEAR);

	XMapWindow (pDisplay, window);

	/* redirect window to offscreen-storage */
	XCompositeRedirectWindow (pDisplay,
				  window,
				  CompositeRedirectAutomatic);

	init_tfp (pDisplay, DefaultScreen (pDisplay));

	event_loop (pDisplay, window);

	XCompositeUnredirectWindow (pDisplay,
				    window,
				    CompositeRedirectAutomatic);

	return 0;
}

